#pragma warning( disable : 4996 )
#pragma warning( disable : 4267 )

#include <Kernel_Simplex.h>


double vxa::simplex_dist(const Array<double>& x1, const Array<double>& x2)
{
  unsigned int N = x1.nelem();

  double d=0;
  for(unsigned int n=0; n < N; ++n)
  {
    d += ::sqrt(x1(n)*x2(n));
  }
  d = Shark::min(1.0,d); // must be in [0,1]
  d = ::acos(d);
  return d;
}

double vxa::simplex_dist(const Array<double>& x1, const Array<double>& x2,
                         const Array<double>& w /*, Array<double>& tmp1, Array<double>& tmp2 */)
{
  unsigned int N = x1.nelem();
  double d=0;

  double x1p=0,x2p=0;
  for(unsigned int n=0; n < N; ++n)
  {
    x1p += x1(n)*w(n);
    x2p += x2(n)*w(n);
  }
  for(unsigned int n=0; n < N; ++n)
  {
    d += ::sqrt(x1(n)*w(n)*x2(n)*w(n)/(x1p*x2p));
  }
  d = Shark::min(d,1.0);
  return ::acos(d);
}


double vxa::kernel_simplex_est_bandwidth(const Array<double>& x)
{
  double mean_dist = 0.0;
  int n=0;
  int N = x.dim(0);
  for(int i1=0; i1 < N; ++i1)
  {
    for(int i2=i1+1; i2 < N; ++i2)
    {
      mean_dist += simplex_dist(x[i1],x[i2]);
      ++n;
    }
  }
  return mean_dist/n;
}

void vxa::KernelSimplex_fix_gamma::setParameter(unsigned int index, double value)
{
  parameter(index) = value;
  //w(index) = value * value;
  w(index) = ::exp(value);
}

void vxa::KernelSimplex::setParameter(unsigned int index, double value)
{
  parameter(index) = value;
  //w(index) = value * value;
  if(index > 0)
  {
    w(index-1) = ::exp(value);
  }
  else
  {
    g = ::exp(value);
  }
}



double vxa::KernelSimplex_fix_gamma::eval
(const Array<double>& x1, const Array<double>& x2) const
{
  unsigned int N = x1.nelem();

  double d = simplex_dist(x1,x2,w);

  double k = exp(-0.5*vxa::Core::sq(d/g));
  return k;
}


double vxa::KernelSimplex_fix_gamma::evalDerivative
(const Array<double>& x1, const Array<double>& x2, Array<double>& drv) const
{
  const double tol = 1e-14;
  unsigned int N = x1.nelem();
  drv.resize(N,false);

  double x1_lam=0,x2_lam=0;
  Array<double> a; a.resize(1,N,false);

  for(unsigned int n=0; n < N; ++n)
  {
    x1_lam += x1(n)*w(n);
    x2_lam += x2(n)*w(n);
    a(n) = ::sqrt(x1(n)*x2(n));
  }

  double z = 0.0;
  for(unsigned int n=0; n < N; ++n)
  {
    z += w(n)*a(n);
  }
  z /= ::sqrt(x1_lam * x2_lam);
  z = Shark::min(1.0,z);
  double d = ::acos(z);

  if(1.0-z < tol) // z is 1.0
  {
    std::fill(drv.begin(),drv.end(),0.0);
    return 1.0;
  }
  double k = ::exp(-0.5*vxa::Core::sq(d/g));

  Array<double> dz; dz.resize(1,N,false);
  double wa_sum = 0.0;
  for(unsigned int j=0; j < N; ++j)
  {
    wa_sum += w(j)*a(j);
  }

  for(unsigned int n=0; n < N; ++n)
  {
    double T4 = wa_sum * (x2(n)*x1_lam + x1(n)*x2_lam);
    dz(n) = 1.0/sqrt(x1_lam*x2_lam) * (a(n) - T4/(2*x1_lam*x2_lam));

    dz(n) *= k * 1.0/(g*g) * d /sqrt(1-z*z);

    //drv(n) = dz(n) * 2*parameter(n); // w(n) = parameter(n)^2
    drv(n) = dz(n) * parameter(n); // w(n) = exp(parameter(n))

  } // der w.r.t. lambda_n
  
  

  return k;
}


#if 0
/*
compare to this initial implementation with N^2 complexity:
*/
double vxa::KernelSimplex_fix_gamma::evalDerivative
(const Array<double>& x1, const Array<double>& x2, Array<double>& drv) const
{
  const double tol = 1e-14;
  unsigned int N = x1.nelem();
  drv.resize(N,false);

  double x1_lam=0,x2_lam=0;
  Array<double> a; a.resize(1,N,false);

  for(unsigned int n=0; n < N; ++n)
  {
    x1_lam += x1(n)*w(n);
    x2_lam += x2(n)*w(n);
    a(n) = ::sqrt(x1(n)*x2(n));
  }

  double z = 0.0;
  for(unsigned int n=0; n < N; ++n)
  {
    z += w(n)*a(n);
  }
  z /= ::sqrt(x1_lam * x2_lam);
  z = Shark::min(1.0,z);
  double d = ::acos(z);

  if(1.0-z < tol) // z is 1.0
  {
    std::fill(drv.begin(),drv.end(),0.0);
    return 1.0;
  }
  double k = ::exp(-0.5*vxa::Core::sq(d/g));

  Array<double> dz; dz.resize(1,N,false);
  for(unsigned int n=0; n < N; ++n)
  {
    double T4=0;
    for(unsigned int j=0; j < N; ++j)
    {
      T4 += w(j)*a(j)*(x2(n)*x1_lam + x1(n)*x2_lam);
    }
    dz(n) = 1.0/sqrt(x1_lam*x2_lam) * (a(n) - T4/(2*x1_lam*x2_lam));

    dz(n) *= k * 1.0/(g*g) * d /sqrt(1-z*z);

    //drv(n) = dz(n) * 2*parameter(n); // w(n) = parameter(n)^2
    drv(n) = dz(n) * parameter(n); // w(n) = exp(parameter(n))

  } // der w.r.t. lambda_n
  return k;
}
#endif



void vxa::test_kernel_simplex_fix_gamma(void)
{
  Array<double> X;
  X.resize(3,2,false);
  double _X[] = {0.1, 0.9, 0.101, 0.899, 0.2, 0.8};
  std::copy(_X,_X+X.nelem(),X.begin());

  vxa::KernelSimplex_fix_gamma k(X.dim(1),0.5);
  k.setParameter(0,log(0.3));
  k.setParameter(1,log(0.7));

  Array<double> der;
  double kv;
  for(size_t i=0; i < X.dim(0); ++i)
  {
    kv = k.evalDerivative(X[0],X[i],der);
    std::cout << "kv=" << kv << " drv=";
    for(size_t j=0; j < der.nelem(); ++j)
    {
      std::cout << der(j) << " ";
    }
    std::cout << std::endl;
    int stophere=1;
  }
}
